# builders.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Common functionality across API versions to build commands, settings, and statuses"""

from __future__ import annotations
import enum
import logging
from datetime import datetime
from pathlib import Path
from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import (
    Any,
    ClassVar,
    TypeVar,
    Generic,
    Type,
    Union,
    no_type_check,
    Optional,
    Dict,
    Callable,
    Tuple,
    List,
)

import wrapt
import requests
import betterproto
from construct import Int8ub, Int16ub, Struct, Adapter, GreedyBytes

from open_gopro.responses import (
    BytesParser,
    BytesBuilder,
    BytesParserBuilder,
    JsonParser,
    GoProResp,
    StringBuilder,
)
from open_gopro.constants import (
    ActionId,
    UUID,
    CmdId,
    ResponseType,
    SettingId,
    QueryCmdId,
    StatusId,
    ErrorCode,
)
from open_gopro.communication_client import GoProBle, GoProWifi

logger = logging.getLogger(__name__)

SettingValueType = TypeVar("SettingValueType")
CommandValueType = TypeVar("CommandValueType")

####################################################### General ##############################################


@wrapt.decorator
# pylint: disable = E, W
def log_query(
    wrapped: Callable, instance: Union[BleSetting, BleStatus, WifiSetting], args: Any, kwargs: Any
) -> Callable:
    """Log a query write."""
    logger.info(f"<----------- {wrapped.__name__} : {instance.identifier}")
    response = wrapped(*args, **kwargs)
    logger.info(f"-----------> {response}")
    return response


######################################################## BLE #################################################


def build_enum_adapter(target: Type[enum.Enum]) -> Adapter:
    """Build an enum to Construct parsing and building adapter

    This adapter only works on byte data of length 1

    Args:
        target (Type[enum.Enum]): Enum to use use for parsing / building

    Returns:
        Adapter: adapter to be used by Construct
    """

    class EnumByteAdapter(Adapter):
        """An enum to Construct adapter"""

        target: ClassVar[Type[enum.Enum]]

        def _decode(self, obj: bytearray, *_: Any) -> enum.Enum:
            """Parse a bytestream of length 1 into an Enum

            Args:
                obj (bytearray): bytestream to parse

            Returns:
                enum.Enum: Enum value
            """
            return self.target(obj)

        def _encode(self, obj: Union[enum.Enum, int], *_: Any) -> int:
            """Adapt an enum for use by Construct

            Args:
                obj (Union[enum.Enum, int]): Enum to adapt

            Returns:
                int: int value of Enum
            """
            return obj if isinstance(obj, int) else obj.value

    setattr(EnumByteAdapter, "target", target)
    return EnumByteAdapter(Int8ub)


status_struct = Struct("status" / build_enum_adapter(ErrorCode))


class DateTimeAdapter(Adapter):
    """Translate between different date time representations"""

    def _decode(self, obj: Union[List, str], *_: Any) -> datetime:
        """Translate string or list of bytes into datetime

        Args:
            obj (list): input

        Returns:
            datetime: built datetime
        """
        if isinstance(obj, str):
            # comes as '%14%01%02%03%09%2F'
            year, *remaining = [int(x, 16) for x in obj.split("%")[1:]]
            return datetime(year + 2000, *remaining)  # type: ignore
        if isinstance(obj, list):
            # When received from BLE, it includes garbage first byte
            obj = obj[-7:]
            year = Int16ub.parse(bytes(obj[0:2]))
            return datetime(year, *[int(x) for x in obj[2:]])  # type: ignore
        raise TypeError(f"Datetime decoder received invalid type: {type(obj)}")

    def _encode(self, obj: Union[datetime, str], *_: Any) -> Union[bytes, str]:
        """Translate datetime into bytes or pass through string

        Args:
            obj (Union[datetime, str]): Input

        Returns:
            Union[bytes, str]: built bytes
        """
        if isinstance(obj, datetime):
            year = [int(x) for x in Int16ub.build(obj.year)]
            return bytes([*year, obj.month, obj.day, obj.hour, obj.minute, obj.second])
        if isinstance(obj, str):
            return obj
        raise TypeError(f"Datetime encoder received invalid type: {type(obj)}")


# Ignoring because hitting this mypy bug: https://github.com/python/mypy/issues/5374
@dataclass  # type: ignore
class BleCommand(ABC):
    """The base class for all BLE commands to store common info

    Args:
        communicator (GoProBle): BLE client to read / write
        uuid (UUID): UUID to read / write to
    """

    communicator: GoProBle
    uuid: UUID

    def __post_init__(self) -> None:
        self.communicator._add_parser(self._identifier, self._response_parser)

    @property
    @abstractmethod
    def _identifier(self) -> ResponseType:
        raise NotImplementedError

    @property
    @abstractmethod
    def _response_parser(self) -> BytesParser:
        raise NotImplementedError


@dataclass
class BleReadCommand(BleCommand):
    """A BLE command that reads data from a UUID

    Args:
        communicator (GoProBle): BLE client to read
        uuid (UUID): UUID to read to
        response_parser (BytesParser): the parser that will parse the received bytestream into a JSON dict
    """

    response_parser: BytesParser

    @property
    def _identifier(self) -> ResponseType:
        return self.uuid

    @property
    def _response_parser(self) -> BytesParser:
        return self.response_parser

    # pylint: disable=missing-return-doc
    def __call__(self) -> GoProResp:  # noqa: D102
        logger.info(f"<----------- {self.uuid.name}")
        response = self.communicator._read_characteristic(self.uuid)
        logger.info(f"-----------> {self.uuid.name} : {response}")
        return response


@dataclass
class BleWriteNoParamsCommand(BleCommand):
    """A BLE command that writes to a UUID and does not accept any parameters

    Args:
        communicator (GoProBle): BLE client to write
        uuid (UUID): UUID to write to
        cmd (CmdId): Command ID that is being sent
        response_parser (Optional[ConstructBytesParser]): the parser that will parse the received bytestream into a JSON dict.
            Defaults to None
    """

    cmd: CmdId
    response_parser: Optional[BytesParser] = None

    @property
    def _identifier(self) -> ResponseType:
        return self.cmd

    @property
    def _response_parser(self) -> BytesParser:
        return status_struct if self.response_parser is None else status_struct + self.response_parser

    # pylint: disable=missing-return-doc
    def __call__(self) -> GoProResp:  # noqa: D102
        logger.info(f"<----------- {self.cmd.name}")
        # Build data buffer
        data = bytearray([self.cmd.value])
        data.insert(0, len(data))
        # Send the data and receive the response
        response = self.communicator._write_characteristic_receive_notification(self.uuid, data)
        logger.info(f"-----------> \n{response}")
        return response


@dataclass
class BleWriteWithParamsCommand(BleCommand, Generic[CommandValueType]):
    """A BLE command that writes to a UUID and does not accept any parameters

    Args:
        communicator (GoProBle): BLE client to write
        uuid (UUID): UUID to write to
        cmd (CmdId): Command ID that is being sent
        param_builder (BytesBuilder): is responsible for building the bytestream to send from the input params
        response_parser (Optional[BytesParser]): the parser that will parse the received bytestream into a JSON dict
    """

    cmd: CmdId
    param_builder: BytesBuilder
    response_parser: Optional[BytesParser] = None

    @property
    def _identifier(self) -> ResponseType:
        return self.cmd

    @property
    def _response_parser(self) -> BytesParser:
        return status_struct if self.response_parser is None else status_struct + self.response_parser

    # pylint: disable=missing-return-doc
    def __call__(self, value: CommandValueType) -> GoProResp:  # noqa: D102
        logger.info(f"<----------- {self.cmd.name}: {str(value)}")
        # Build data buffer
        data = bytearray([self.cmd.value])
        # Mypy is not understanding the subclass check here
        param = value.value if issubclass(type(value), enum.Enum) else value  # type: ignore
        # There must be a param builder if we have a param
        param = self.param_builder.build(param)
        data.extend([len(param), *param])
        data.insert(0, len(data))
        # Send the data and receive the response
        response = self.communicator._write_characteristic_receive_notification(self.uuid, data)
        logger.info(f"-----------> \n{response}")
        return response


@dataclass
class RegisterUnregisterAll(BleWriteNoParamsCommand):
    """Base class for register / unregister all commands

    This will loop over all of the elements (i.e. settings / statusess found from the element_set entry of the
    producer tuple parameter) and individually register / unregister (depending on the action parameter) each
    element in the set

    Args:
        producer: (Optional[Tuple[Union[Type[SettingId], Type[StatusId]], QueryCmdId]]): Tuple of (element_set,
            query command) where element_set is the GoProEnum that this command relates to, i.e. SettingId for
            settings, StatusId for Statuses
        action: (Optional[Action]): whether to register or unregister
    """

    class Action(enum.Enum):
        """Enum to differentiate between register actions"""

        REGISTER = enum.auto()
        UNREGISTER = enum.auto()

    producer: Optional[Tuple[Union[Type[SettingId], Type[StatusId]], QueryCmdId]] = None
    action: Optional[Action] = None

    def __post_init__(self) -> None:
        # TODO refactor to not use dataclasses since derived classes can't have non default members if base classes do
        assert self.producer is not None
        assert self.action is not None

    # pylint: disable=missing-return-doc
    def __call__(self) -> GoProResp:  # noqa: D102
        assert self.producer is not None
        assert self.action is not None
        element_set = self.producer[0]
        responded_command = self.producer[1]
        response = super().__call__()
        if response.is_ok:
            for element in element_set:
                (
                    self.communicator._register_listener
                    if self.action is RegisterUnregisterAll.Action.REGISTER
                    else self.communicator._unregister_listener
                )(
                    # Ignoring typing because this seems correct and looks like mypy error
                    (responded_command, element)  # type: ignore
                )

        return response


def build_protobuf_adapter(protobuf: Type[betterproto.Message]) -> Adapter:
    """Build a protobuf to Construct parsing (only) adapter

    Args:
        protobuf (Type[betterproto.Message]): protobuf to use as parser

    Returns:
        Adapter: adapter to be used by Construct
    """

    class ProtobufConstructAdapter(Adapter):
        """Adapt a protobuf to be used by Construct (for parsing only)"""

        protobuf: Type[betterproto.Message]

        def _decode(self, obj: bytearray, *_: Any) -> Dict[Any, Any]:
            """Parse a byte stream into a JSON dict using a protobuf

            Args:
                obj (bytearray): byte stream to parse

            Returns:
                Dict[Any, Any]: parsed JSON dict
            """
            return self.protobuf.FromString(bytes(obj)).to_dict()

        def _encode(self, *_: Any) -> Any:
            raise NotImplementedError

    setattr(ProtobufConstructAdapter, "protobuf", protobuf)
    return ProtobufConstructAdapter(GreedyBytes)


# Ignoring because hitting this mypy bug: https://github.com/python/mypy/issues/5374
@dataclass  # type: ignore
class BleProtoCommand(BleCommand):
    """A BLE command that writes to a UUID and does not accept any parameters

    Args:
        communicator (GoProBle): BLE client to write
        uuid (UUID): UUID to write to
        feature_id (CmdId): Command ID that is being sent
        action_id (ActionId): protobuf specific action ID that is being sent
        request_proto (Type[betterproto.Message]): protobuf used to build command bytestream
        response_proto (Type[betterproto.Message]): protobuf used to parse received bytestream
    """

    feature_id: CmdId
    action_id: ActionId
    request_proto: Type[betterproto.Message]
    response_proto: Type[betterproto.Message]

    @property
    def _identifier(self) -> ResponseType:
        return self.action_id

    @property
    def _response_parser(self) -> BytesParser:
        return (
            self.response_proto if self.response_proto is None else build_protobuf_adapter(self.response_proto)
        )

    @abstractmethod
    @no_type_check
    # pylint: disable=missing-return-doc
    def __call__(self, *args: Any, **kwargs: Any) -> GoProResp:  # noqa: D102
        # The method that will actually build and send the protobuf command

        # This method's signature shall be override by the subclass.
        # The subclass shall then pass the arguments to this method and return it's returned response

        # This pattern is technically violating the Liskov substitution principle. But we are accepting this as a
        # tradeoff for exposing type hints on BLE Protobuf commands.
        logger.info(
            f"<----------- {self.feature_id.name} : {' '.join([*[str(a) for a in args], *[str(a) for a in kwargs.values()]])}"
        )

        # Build request protobuf bytestream
        proto = self.request_proto()
        # Add args to protobuf request
        attrs = iter(self.__call__.__annotations__.keys())
        for arg in args:
            param = arg.value if issubclass(type(arg), enum.Enum) else arg
            setattr(proto, next(attrs), param)
        # Add kwargs to protobuf request
        for name, arg in kwargs.items():
            if arg is not None:
                param = arg.value if issubclass(type(arg), enum.Enum) else arg
                setattr(proto, name, param)

        # Prepend headers and serialize
        request = bytearray([self.feature_id.value, self.action_id.value, *proto.SerializeToString()])
        # Prepend length
        request.insert(0, len(request))

        # Allow exception to pass through if protobuf not completely initialized
        response = self.communicator._write_characteristic_receive_notification(self.uuid, request)
        logger.info(f"-----------> \n{response}")
        return response


class BleSetting(Generic[SettingValueType]):
    """An individual camera setting that is interacted with via BLE.

    Args:
        communicator (GoProBle): Adapter to read / write settings data
        identifier (SettingId): ID of setting
        parser_builder (BytesParserBuilder): object to both parse and build setting
    """

    def __init__(
        self, communicator: GoProBle, identifier: SettingId, parser_builder: BytesParserBuilder
    ) -> None:
        self.identifier = identifier
        self.communicator: GoProBle = communicator
        self.setter_uuid: UUID = UUID.CQ_SETTINGS
        self.reader_uuid: UUID = UUID.CQ_QUERY
        self.parser: BytesParser = parser_builder
        self.builder: BytesBuilder = parser_builder  # Just syntactic sugar
        communicator._add_parser(self.identifier, self.parser)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return str(self.identifier.name)

    def set(self, value: SettingValueType) -> GoProResp:
        """Set the value of the setting.

        Args:
            value (Any): The argument to use to set the setting value.

        Returns:
            GoProResp: Status of set
        """
        logger.info(f"<----------- Set {self.identifier.name}: {str(value)}")

        data = bytearray([self.identifier.value])
        try:
            param = self.builder.build(value)
            data.extend([len(param), *param])
        except IndexError:
            pass
        data.insert(0, len(data))
        response = self.communicator._write_characteristic_receive_notification(self.setter_uuid, data)

        logger.info(f"-----------> \n{response}")
        return response

    @log_query
    def get_value(self) -> GoProResp:
        """Get the settings value.

        Returns:
            GoProResp: settings value
        """
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.GET_SETTING_VAL)
        )

    @log_query
    def get_name(self) -> GoProResp:
        """Get the settings name.

        Raises:
            NotImplementedError: This isn't implemented on the camera

        Returns:
            GoProResp: settings name
        """
        # return self.communicator._write_characteristic_receive_notification(
        #     self.reader_uuid, QueryCmdId.GET_SETTING_NAME, self._build_cmd(QueryCmdId.GET_SETTING_NAME)
        # )
        raise NotImplementedError("Not implemented on camera!")

    @log_query
    def get_capabilities_values(self) -> GoProResp:
        """Get currently supported settings capabilities values.

        Returns:
            GoProResp: settings capabilities values
        """
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.GET_CAPABILITIES_VAL)
        )

    @log_query
    def get_capabilities_names(self) -> GoProResp:
        """Get currently supported settings capabilities names.

        Raises:
            NotImplementedError: This isn't implemented on the camera

        Returns:
            GoProResp: settings capabilities names
        """
        # return self.communicator._write_characteristic_receive_notification(
        #     self.reader_uuid,
        #     QueryCmdId.GET_CAPABILITIES_NAME,
        #     self._build_cmd(QueryCmdId.GET_CAPABILITIES_NAME),
        # )
        raise NotImplementedError("Not implemented on camera!")

    @log_query
    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Current value of respective setting ID
        """
        self.communicator._register_listener((QueryCmdId.SETTING_VAL_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.REG_SETTING_VAL_UPDATE)
        )

    @log_query
    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator._unregister_listener((QueryCmdId.SETTING_VAL_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.UNREG_SETTING_VAL_UPDATE)
        )

    @log_query
    def register_capability_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's capabilities update.

        Returns:
            GoProResp: Current capabilities of respective setting ID
        """
        self.communicator._register_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.REG_CAPABILITIES_UPDATE)
        )

    @log_query
    def unregister_capability_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's capabilities change.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator._unregister_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            self.reader_uuid, self._build_cmd(QueryCmdId.UNREG_CAPABILITIES_UPDATE)
        )

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data to send a settings query over-the-air.

        Args:
            cmd (QueryCmdId): command to build

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self.identifier.value])
        ret.insert(0, len(ret))
        return ret


class BleStatus:
    """An individual camera status that is interacted with via BLE.

    Args:
        communicator (GoProBle): Adapter to read status data
        identifier (StatusId): ID of status
    """

    uuid = UUID.CQ_QUERY

    def __init__(self, communicator: GoProBle, identifier: StatusId, parser: BytesParser) -> None:
        self.identifier = identifier
        self.communicator = communicator
        self.parser = parser
        # Add to response parsing map
        communicator._add_parser(self.identifier, self.parser)

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return str(self.identifier.name)

    @log_query
    def get_value(self) -> GoProResp:
        """Get the current value of a status.

        Returns:
            GoProResp: current status value
        """
        return self.communicator._write_characteristic_receive_notification(
            BleStatus.uuid, self._build_cmd(QueryCmdId.GET_STATUS_VAL)
        )

    @log_query
    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a status changes.

        Returns:
            GoProResp: current status value
        """
        self.communicator._register_listener((QueryCmdId.STATUS_VAL_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            BleStatus.uuid, self._build_cmd(QueryCmdId.REG_STATUS_VAL_UPDATE)
        )

    @log_query
    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when status changes.

        Returns:
            GoProResp: Status of unregister
        """
        self.communicator._unregister_listener((QueryCmdId.STATUS_VAL_PUSH, self.identifier))
        return self.communicator._write_characteristic_receive_notification(
            BleStatus.uuid, self._build_cmd(QueryCmdId.UNREG_STATUS_VAL_UPDATE)
        )

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data for a given status command.

        Args:
            cmd (QueryCmdId): command to build data for

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self.identifier.value])
        ret.insert(0, len(ret))
        return ret


######################################################## Wifi #################################################


@dataclass
class WifiGetJsonCommand:
    """The base class for all WiFi Commands. Stores common information.

    Args:
        communicator (GoProWifi): Wifi client to write command
        endpoint (str): endpoint to GET
    """

    communicator: GoProWifi
    endpoint: str
    response_parser: Optional[JsonParser] = None

    def __post_init__(self) -> None:
        if self.response_parser is not None:
            self.communicator._add_parser(self.endpoint, self.response_parser)


@dataclass
class WifiGetJsonWithParams(WifiGetJsonCommand, Generic[CommandValueType]):
    """A Wifi command that writes to a UUID (with parameters) and receives JSON as response

    Args:
        communicator (GoProWifi): Wifi client to write command
        endpoint (str): endpoint to GET
        response_parser (Optional[JsonParser]): the parser that will parse the received bytestream into a JSON dict
    """

    communicator: GoProWifi
    endpoint: str
    response_parser: Optional[JsonParser] = None
    param_builder: Optional[StringBuilder] = None

    # pylint: disable=missing-return-doc
    def __call__(self, value: CommandValueType) -> GoProResp:  # noqa: D102
        logger.info(f"<----------- {self.endpoint} : {str(value)}")

        # Build list of args as they should be represented in URL
        url_params = []
        if self.param_builder is not None:
            url_params.append(self.param_builder(value))
        elif issubclass(type(value), enum.Enum):
            url_params.append(value.value)  # type: ignore
        else:
            url_params.append(value)  # type: ignore
        url = self.endpoint.format(*url_params)
        # Send to camera
        response = self.communicator._get(url)
        logger.info(f"-----------> \n{response}")
        return response


@dataclass
class WifiGetJsonNoParams(WifiGetJsonCommand):
    """A Wifi command that writes to a UUID (with parameters) and receives JSON as response

    Args:
        communicator (GoProWifi): Wifi client to write command
        endpoint (str): endpoint to GET
        response_parser (Optional[JsonParser]): the parser that will parse the received bytestream into a JSON dict
    """

    # pylint: disable=missing-return-doc
    def __call__(self) -> GoProResp:  # noqa: D102
        logger.info(f"<----------- {self.endpoint}")

        url = self.endpoint
        # Send to camera
        response = self.communicator._get(url)
        logger.info(f"-----------> \n{response}")
        return response


# Ignoring because hitting this mypy bug: https://github.com/python/mypy/issues/5374
@dataclass  # type: ignore
class WifiGetBinary(ABC):
    """A Wifi command that writes to a UUID (with parameters) and receives a binary stream as response

    Args:
        communicator (GoProWifi): Wifi client to write command
        endpoint (str): endpoint to GET
    """

    communicator: GoProWifi
    endpoint: str

    @abstractmethod
    @no_type_check
    # pylint: disable=missing-return-doc
    def __call__(self, /, **kwargs) -> Path:  # noqa: D102
        # The method that will actually send the command and receive the stream

        # This method's signature shall be override by the subclass.
        # The subclass shall then pass the arguments to this method and return it's returned response

        # This pattern is technically violating the Liskov substitution principle. But we are accepting this as a
        # tradeoff for exposing type hints on BLE Protobuf commands.
        logger.info(f'<----------- {self.endpoint} : {" ".join([str(x) for x in list(kwargs.values())])}')
        camera_file = kwargs["camera_file"]
        try:
            local_file = Path(kwargs["local_file"])
        except KeyError:
            local_file = Path(".") / camera_file

        url = self.endpoint.format(camera_file)
        # Send to camera
        try:
            self.communicator._stream_to_file(url, local_file)
        except requests.exceptions.HTTPError as e:
            logger.error(repr(e))
        else:
            logger.info("-----------> SUCCESS")
        return local_file


class WifiSetting(Generic[SettingValueType]):
    """An individual camera setting that is interacted with via Wifi.

    Args:
        communicator (WifiCommunicator): Adapter to read / write settings data
        id (SettingId): ID of setting
        endpoint (str): HTTP endpoint to be used to set setting value
    """

    def __init__(self, communicator: GoProWifi, identifier: SettingId) -> None:
        self.identifier = identifier
        self.communicator = communicator
        # Note! It is assumed that BLE and WiFi settings are symmetric so we only add to the communicator's
        # parser in the BLE Setting.

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return str(self.identifier.name)

    def set(self, value: SettingValueType) -> GoProResp:
        """Set the value of the setting.

        Args:
            value (Any): value to set setting

        Returns:
            GoProResp: Status of set
        """
        logger.info(f"<----------- Setting {self.identifier}: {value}")
        # Build url. TODO fix this type error with Mixin (or passing in endpoint as argument)
        value = value.value if isinstance(value, enum.Enum) else value
        url = self.communicator._api.wifi_setting.endpoint.format(self.identifier.value, value)  # type: ignore
        # Send to camera
        response = self.communicator._get(url)
        if response is not None:
            logger.info(f"-----------> \n{response}")
        return response



# util.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:50 PM

"""Miscellaneous utilities for the GoPro package."""

import sys
import queue
import logging
import subprocess
from pathlib import Path
from typing import Dict, Type, Any, List, Optional, Union

from rich.logging import RichHandler
from rich import traceback

util_logger = logging.getLogger(__name__)


def setup_logging(logger: Any, output: Path, modules: Dict[str, int] = None) -> Any:
    """Configure open gopro modules for logging

    The application's logger is passed in, modified, and then returned

    Args:
        logger (Any): input logger that will be modified and then returned
        output (Path): Path of log file for file stream handler
        modules (Dict[str, int], optional): Optional override of modules / levels.

    Returns:
        Any: updated logger that the application can use for logging
    """
    modules = modules or {
        "open_gopro.gopro": logging.DEBUG,
        "open_gopro.api.builders": logging.DEBUG,
        "open_gopro.api.v1_0.wifi_commands": logging.DEBUG,
        "open_gopro.communication_client": logging.DEBUG,
        "open_gopro.ble.adapters.bleak_wrapper": logging.DEBUG,
        "open_gopro.wifi.adapters.wireless": logging.DEBUG,
        "open_gopro.responses": logging.DEBUG,
        "open_gopro.util": logging.DEBUG,
        "bleak": logging.DEBUG,
    }

    # Logging to file with millisecond timing
    fh = logging.FileHandler(output, mode="w")
    file_formatter = logging.Formatter(
        fmt="%(threadName)13s:%(asctime)s.%(msecs)03d %(filename)-40s %(lineno)4s %(levelname)-8s | %(message)s",
        datefmt="%H:%M:%S",
    )
    fh.setFormatter(file_formatter)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    # Use Rich for colorful console logging
    sh = RichHandler(rich_tracebacks=True, enable_link_path=True, show_time=False)
    stream_formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S")
    sh.setFormatter(stream_formatter)
    sh.setLevel(logging.INFO)
    logger.addHandler(sh)
    logger.setLevel(logging.DEBUG)

    # Enable / disable logging in modules
    for module, level in modules.items():
        l = logging.getLogger(module)
        l.setLevel(level)
        l.addHandler(fh)
        l.addHandler(sh)

    traceback.install()  # Enable exception tracebacks in rich logger

    return logger


def set_logging_level(logger: Any, level: int) -> None:
    """Change the global logging level

    Args:
        logger (Any): logger to update
        level (int): level to set
    """
    for handler in logger.handlers:
        if isinstance(handler, RichHandler):
            handler.setLevel(level)


def launch_vlc(location: Optional[Path]) -> None:
    """Launch VLC

    Args:
        location (Optional[Path]): path to VLC. If None, it will be automatically discovered
    """
    # This is a fairly lazy way to find VLC. We'll call it best effort.
    potential_vlc_locations: List[Union[Path, str]] = []
    command = "echo Invalid Platform"
    if "linux" in sys.platform.lower():
        potential_vlc_locations = [r'"/snap/bin/vlc"']
        command = 'su $(id -un 1000) -c "{} udp://@:8554 > /dev/null 2>&1 &"'
    elif "darwin" in sys.platform.lower():
        potential_vlc_locations = [r'"/Applications/VLC.app/Contents/MacOS/VLC"']
        command = "{} udp://@:8554 > /dev/null 2>&1 &"
    elif "win" in sys.platform.lower():
        potential_vlc_locations = [
            r'"/c/Program Files/VideoLAN/VLC/vlc.exe"',
            r'"/c/Program Files (x86)/VideoLAN/VLC/vlc.exe"',
            r'"C:\Program Files (x86)\VideoLAN\VLC\vlc.exe"',
            r'"C:\Program Files\VideoLAN\VLC\vlc.exe"',
        ]
        command = "{} udp://@:8554 &"

    potential_vlc_locations = potential_vlc_locations if location is None else [location]
    for vlc in potential_vlc_locations:
        response = cmd(command.format(vlc)).lower()

        if (
            " not " not in response
            and " no " not in response
            and " cannot " not in response
            and " unexpected " not in response
        ):
            util_logger.info("VLC launched")
            return

    util_logger.error("Failed to find VLC")


def scrub(obj: Any, bad_key: str) -> None:
    """Recursively scrub a dict or list to remove a given key in place.

    Args:
        obj (Any): dict or list to operate on. If neither, it will return immediately.
        bad_key (str): key to remove
    """
    if isinstance(obj, dict):
        for key in list(obj.keys()):
            if key == bad_key:
                del obj[key]
            else:
                scrub(obj[key], bad_key)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            if obj[i] == bad_key:
                del obj[i]
            else:
                scrub(obj[i], bad_key)
    else:
        # neither a dict nor a list, do nothing
        pass


def cmd(command: str) -> str:
    """Send a command to the shell and return the result.

    Args:
        command (str): command to send

    Returns:
        str: response returned from shell
    """
    util_logger.debug(f"Send cmd --> {command}")
    # Note: Ignoring unicode characters in SSIDs to prevent intermittent UnicodeDecodeErrors from occurring
    # while trying to connect to SSID when *any* AP is nearby that has unicode characters in the name
    response = (
        subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # type: ignore
        .stdout.read()
        .decode(errors="ignore")
    )
    util_logger.debug(f"Receive response --> {response}")

    return response


class Singleton:
    """To be subclassed to create a singleton class."""

    _instances: Dict[Type["Singleton"], Type["Singleton"]] = {}

    # pylint: disable=missing-return-doc, missing-return-type-doc
    def __new__(cls, *args: Any, **kwargs: Any) -> Any:
        """Check for existing instance."""
        if cls not in cls._instances:
            # https://github.com/python/mypy/issues/6061
            cls._instances[cls] = object.__new__(cls, *args, **kwargs)  # type: ignore
        return cls._instances[cls]


class SnapshotQueue(queue.Queue):
    """A subclass of the default queue module to safely take a snapshot of the queue

    This is so we can access the elements (in a thread safe manner) without dequeuing them.
    """

    def snapshot(self) -> List[Any]:
        """Acquire the mutex, then return the queue's elements as a list.

        Returns:
            List[Any]: List of queue elements
        """
        with self.mutex:
            return list(self.queue)



# log_battery.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:45 PM

"""Example to continuously read the battery (with no Wifi connection)"""

import csv
import time
import logging
import argparse
import threading
from pathlib import Path
from datetime import datetime
from dataclasses import dataclass
from typing import Optional, Tuple, Literal, List

from rich.console import Console

from open_gopro import GoPro
from open_gopro.constants import StatusId
from open_gopro.util import setup_logging, set_logging_level

logger = logging.getLogger(__name__)
console = Console()  # rich consoler printer

BarsType = Literal[0, 1, 2, 3]


@dataclass
class Sample:
    """Simple class to store battery samples"""

    index: int
    percentage: int
    bars: BarsType

    def __post_init__(self) -> None:
        self.time = datetime.now()

    def __str__(self) -> str:  # pylint: disable=missing-return-doc
        return f"Index {self.index} @ time {self.time.strftime('%H:%M:%S')} --> bars: {self.bars}, percentage: {self.percentage}"


SAMPLE_INDEX = 0
SAMPLES: List[Sample] = []


def dump_results_as_csv(location: Path) -> None:
    """Write all of the samples to a csv file

    Args:
        location (Path): File to write to
    """
    console.print(f"Dumping results as CSV to {location}")
    with open(location, mode="w") as f:
        w = csv.writer(f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL)
        w.writerow(["index", "time", "percentage", "bars"])
        initial_time = SAMPLES[0].time
        for s in SAMPLES:
            w.writerow([s.index, (s.time - initial_time).seconds, s.percentage, s.bars])


def process_battery_notifications(gopro: GoPro, initial_bars: BarsType, initial_percentage: int) -> None:
    """Separate thread to continuously check for and store battery notifications.

    If the CLI parameter was set to poll, this isn't used.

    Args:
        gopro (GoPro): instance to get updates from
        initial_bars (BarsType): Initial bars level when notifications were enabled
        initial_percentage (int): Initial percentage when notifications were enabled
    """
    global SAMPLE_INDEX
    global SAMPLES
    last_percentage = initial_percentage
    last_bars = initial_bars

    while True:
        # Block until we receive an update
        notification = gopro.get_update()
        # Update data points if they have changed
        last_percentage = (
            notification.data[StatusId.INT_BATT_PER]
            if StatusId.INT_BATT_PER in notification.data
            else last_percentage
        )
        last_bars = (
            notification.data[StatusId.BATT_LEVEL] if StatusId.BATT_LEVEL in notification.data else last_bars
        )
        # Append and print sample
        SAMPLES.append(Sample(index=SAMPLE_INDEX, percentage=last_percentage, bars=last_bars))
        console.print(str(SAMPLES[-1]))
        SAMPLE_INDEX += 1


def main() -> int:
    """Main program functionality

    Returns:
        int: program return code
    """
    identifier, log_location, poll = parse_arguments()
    global logger
    logger = setup_logging(logger, log_location)

    global SAMPLE_INDEX
    global SAMPLES

    gopro: Optional[GoPro] = None
    return_code = 0
    try:
        with GoPro(identifier, enable_wifi=False) as gopro:
            set_logging_level(logger, logging.ERROR)

            # # Setup notifications if we are not polling
            if poll is None:
                console.print("Configuring battery notifications...")
                # Enable notifications of the relevant battery statuses. Also store initial values.
                bars = gopro.ble_status.batt_level.register_value_update().flatten
                percentage = gopro.ble_status.int_batt_per.register_value_update().flatten
                # Start a thread to handle asynchronous battery level notifications
                threading.Thread(
                    target=process_battery_notifications, args=(gopro, bars, percentage), daemon=True
                ).start()
                with console.status("[bold green]Receiving battery notifications until it dies..."):
                    # Sleep forever, allowing notification handler thread to deal with battery level notifications
                    while True:
                        time.sleep(1)
            # Otherwise, poll
            else:
                with console.status("[bold green]Polling the battery until it dies..."):
                    while True:
                        SAMPLES.append(
                            Sample(
                                index=SAMPLE_INDEX,
                                percentage=gopro.ble_status.int_batt_per.get_value().flatten,
                                bars=gopro.ble_status.batt_level.get_value().flatten,
                            )
                        )
                        console.print(str(SAMPLES[-1]))
                        SAMPLE_INDEX += 1
                        time.sleep(poll)

    except Exception as e:  # pylint: disable=broad-except
        logger.error(repr(e))
        return_code = 1
    except KeyboardInterrupt:
        logger.warning("Received keyboard interrupt. Shutting down...")
    finally:
        if len(SAMPLES) > 0:
            csv_location = Path(log_location.parent) / "battery_results.csv"
            dump_results_as_csv(csv_location)
        if gopro is not None:
            gopro.close()
        console.print("Exiting...")
        return return_code  # pylint: disable=lost-exception


def parse_arguments() -> Tuple[str, Path, Optional[int]]:
    """Parse command line arguments

    Returns:
        Tuple[str, Path, Path]: (identifier, path to save log, path to VLC)
    """
    parser = argparse.ArgumentParser(
        description="Connect to the GoPro via BLE only and continuously read the battery (either by polling or notifications)."
    )
    parser.add_argument(
        "-i",
        "--identifier",
        type=str,
        help="Last 4 digits of GoPro serial number, which is the last 4 digits of the default camera SSID. \
            If not used, first discovered GoPro will be connected to",
        default=None,
    )
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        help="Location to store detailed log",
        default="log_battery.log",
    )
    parser.add_argument(
        "-p",
        "--poll",
        type=int,
        help="Set to poll the battery at a given interval. If not set, battery level will be notified instead. Defaults to notifications.",
        default=None,
    )
    args = parser.parse_args()

    return args.identifier, args.log, args.poll


if __name__ == "__main__":
    main()



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Open GoPro BLE Interface interace and implementation"""

from open_gopro.exceptions import FailedToFindDevice, ConnectFailed, ConnectionTerminated, ResponseTimeout
from .services import AttributeTable, Characteristic, Descriptor, Service, UUID
from .controller import BleDevice, BleHandle, NotiHandlerType, DisconnectHandlerType, BLEController
from .client import BleClient



# params.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

# pylint: disable=missing-class-docstring

"""Parameter definitions for GoPro BLE and WiFi commands for Open GoPro version 1.0

Note these have to be IntEnum's in order to be correctly built when sending commands
"""

from open_gopro.constants import GoProEnum

# from open_gopro.proto.request_get_preset_status_pb import EnumRegisterPresetStatus


class ParamsV1_0:
    class Shutter(GoProEnum):
        ON = 1
        OFF = 0

    class Preset(GoProEnum):
        ACTIVITY = 1
        BURST_PHOTO = 65538
        CINEMATIC = 2
        LIVE_BURST = 65537
        NIGHT_PHOTO = 65539
        NIGHT_LAPSE = 131074
        PHOTO = 65536
        STANDARD = 0
        TIME_LAPSE = 131073
        TIME_WARP = 131072
        MAX_PHOTO = 262144
        MAX_TIMEWARP = 327680
        MAX_VIDEO = 196608

    class PresetGroup(GoProEnum):
        VIDE0 = 1000
        PHOTO = 1001
        TIMELAPSE = 1002

    class Resolution(GoProEnum):
        NOT_APPLICABLE = 0
        RES_4K = 1
        RES_2_7K = 4
        RES_2_7K_4_3 = 6
        RES_1440 = 7
        RES_1080 = 9
        RES_4K_4_3 = 18
        RES_5_K = 24

    class FPS(GoProEnum):
        FPS_240 = 0
        FPS_120 = 1
        FPS_100 = 2
        FPS_60 = 5
        FPS_50 = 6
        FPS_30 = 8
        FPS_25 = 9
        FPS_24 = 10
        FPS_200 = 13

    class AutoOff(GoProEnum):
        NEVER = 0
        MIN_5 = 4
        MIN_15 = 6
        MIN_30 = 7

    class LensMode(GoProEnum):
        SINGLE = 0
        DUAL = 5

    class VideoFOV(GoProEnum):
        WIDE = 0
        NARROW = 6
        SUPERVIEW = 3
        LINEAR = 4
        MAX_SUPERVIEW = 7
        LINEAR_HORIZON = 8

    class PhotoFOV(GoProEnum):
        NOT_APPLICABLE = 0
        WIDE = 22
        LINEAR = 23
        NARROW = 24
        MAX_SUPERVIEW = 25

    class MultishotFOV(GoProEnum):
        NOT_APPLICABLE = 0
        WIDE = 22
        LINEAR = 23
        NARROW = 24

    class LED(GoProEnum):
        ALL_OFF = 0
        FRONT_OFF = 1
        ALL_ON = 2
        BLE_KEEP_ALIVE = 66

    class PairState(GoProEnum):
        SUCCESS = 0
        IN_PROGRESS = 1
        FAILED = 2
        STOPPED = 3

    class PairType(GoProEnum):
        NOT_PAIRING = 0
        PAIRING_APP = 1
        PAIRING_REMOTE_CONTROL = 2
        PAIRING_BLUETOOTH = 3

    class WAPState(GoProEnum):
        NEVER_STARTED = 0
        STARTED = 1
        ABORTED = 2
        CANCELED = 3
        COMPLETED = 4

    class SDStatus(GoProEnum):
        OK = 0
        FULL = 1
        REMOVED = 2
        FORMAT_ERROR = 3
        BUSY = 4
        SWAPPED = 8
        UNKNOWN = 0xFF

    class OTAStatus(GoProEnum):
        IDLE = 0
        DOWNLOADING = 1
        VERIFYING = 2
        DOWNLOAD_FAILED = 3
        VERIFY_FAILED = 4
        READY = 5
        GOPRO_APP_DOWNLOADING = 6
        GOPRO_APP_VERIFYING = 7
        GOPRO_APP_DOWNLOAD_FAILED = 8
        GOPRO_APP_VERIFY_FAILED = 9
        GOPRO_APP_READY = 10

    class AnalyticsState(GoProEnum):
        NOT_READY = 0
        READY = 1
        ON_CONNECT = 2

    class ExposureMode(GoProEnum):
        DISABLED = 0
        AUTO = 1
        ISO_LOCK = 2
        HEMISPHERE = 3

    class AccMicStatus(GoProEnum):
        NOT_CONNECTED = 0
        CONNECTED = 1
        CONNECTED_AND_PLUGGED_IN = 2

    class WifiBand(GoProEnum):
        BAND_2_4_GHZ = 0
        BAND_5_GHZ = 1
        BAND_MAX = 2

    class Orientation(GoProEnum):
        UPRIGHT = 0
        UPSIDE_DOWN = 1
        ON_RIGHT_SIDE = 2
        ON_LEFT_SIDE = 3

    class MediaModMicStatus(GoProEnum):
        REMOVED = 0
        ONLY = 1
        WITH_EXTERNAL = 2

    class TimeWarpSpeed(GoProEnum):
        SPEED_15X = 0
        SPEED_30X = 1
        SPEED_60X = 2
        SPEED_150X = 3
        SPEED_300X = 4
        SPEED_900X = 5
        SPEED_1800X = 6
        SPEED_2X = 7
        SPEED_5X = 8
        SPEED_10X = 9
        SPEED_AUTO = 10
        SPEED_1X = 11
        SPEED_HALF = 12

    class MaxLensMode(GoProEnum):
        DEFAULT = 0
        MAX_LENS = 1

    class MediaModStatus(GoProEnum):
        SELFIE_0_HDMI_0_MEDIAMODCONNECTED_FALSE = 0
        SELFIE_0_HDMI_0_MEDIAMODCONNECTED_TRUE = 1
        SELFIE_0_HDMI_1_MEDIAMODCONNECTED_FALSE = 2
        SELFIE_0_HDMI_1_MEDIAMODCONNECTED_TRUE = 3
        SELFIE_1_HDMI_0_MEDIAMODCONNECTED_FALSE = 4
        SELFIE_1_HDMI_0_MEDIAMODCONNECTED_TRUE = 5
        SELFIE_1_HDMI_1_MEDIAMODCONNECTED_FALSE = 6
        SELFIE_1_HDMI_1_MEDIAMODCONNECTED_TRUE = 7

    class Flatmode(GoProEnum):
        VIDEO = 12
        LOOPING = 15
        SINGLE_PHOTO = 16
        NIGHT_PHOTO = 18
        BURST_PHOTO = 19
        TIME_LAPSE_VIDEO = 13
        TIME_LAPSE_PHOTO = 20
        NIGHT_LAPSE_PHOTO = 21
        TIME_WARP_VIDEO = 24
        LIVE_BURST = 25
        NIGHT_LAPSE_VIDEO = 26
        SLO_MO = 27

    class Toggle(GoProEnum):
        ENABLE = 1
        DISABLE = 0
